@@ -28,6 +28,7 @@
 #include <X11/Xutil.h>
 #include <X11/extensions/XTest.h>
 #include <X11/extensions/Xinerama.h>
+#include <X11/extensions/XInput2.h>
 #include <X11/keysym.h>
 #include <X11/cursorfont.h>
 
@@ -69,6 +70,7 @@ static int _xdo_mousebutton(const xdo_t *xdo, Window window, int button, int is_
 static int _is_success(const char *funcname, int code, const xdo_t *xdo);
 static void _xdo_debug(const xdo_t *xdo, const char *format, ...);
 static void _xdo_eprintf(const xdo_t *xdo, int hushable, const char *format, ...);
+static int appears_to_be_wayland(Display *xdpy);
 
 /* context-free functions */
 static wchar_t _keysym_to_char(KeySym keysym);
@@ -83,14 +85,23 @@ static Atom atom_UTF8_STRING = -1;
 xdo_t* xdo_new(const char *display_name) {
   Display *xdpy;
 
-  if ((xdpy = XOpenDisplay(display_name)) == NULL) {
-    /* Can't use _xdo_eprintf yet ... */
-    fprintf(stderr, "Error: Can't open display: %s\n", display_name);
+  if (display_name == NULL) {
+    display_name = XDisplayName(display_name);
+  }
+
+#define DISPLAY_HINT "Is there an Xorg or other X server running? You can try setting 'export DISPLAY=:0' and trying again."
+  if (display_name == NULL) {
+    fprintf(stderr, "Error: No DISPLAY environment variable is set. " DISPLAY_HINT "\n");
     return NULL;
   }
 
-  if (display_name == NULL) {
-    display_name = getenv("DISPLAY");
+  if (*display_name == '\0') {
+    fprintf(stderr, "Error: DISPLAY environment variable is empty. " DISPLAY_HINT "\n");
+    return NULL;
+  }
+
+  if ((xdpy = XOpenDisplay(display_name)) == NULL) {
+    return NULL;
   }
 
   return xdo_new_with_opened_display(xdpy, display_name, 1);
@@ -106,6 +117,14 @@ xdo_t* xdo_new_with_opened_display(Display *xdpy, const char *display,
     return NULL;
   }
 
+  // This library and xdotool do not work correctly on Wayland/XWayland.
+  // Try to detect XWayland and warn the user about problems.
+  if (appears_to_be_wayland(xdpy)) {
+    fprintf(stderr, "The X server at %s appears to be XWayland. Unfortunately, XWayland does not correctly support the features used by libxdo and xdotool.\n", display);
+    return NULL;
+  }
+
+
   /* XXX: Check for NULL here */
   xdo = malloc(sizeof(xdo_t));
   memset(xdo, 0, sizeof(xdo_t));
@@ -212,7 +231,7 @@ int xdo_get_window_location(const xdo_t *xdo, Window wid,
       y = attr.y;
     } else {
       XTranslateCoordinates(xdo->xdpy, wid, attr.root,
-                            attr.x, attr.y, &x, &y, &unused_child);
+                            0, 0, &x, &y, &unused_child);
     }
 
     if (x_ret != NULL) {
@@ -370,7 +389,7 @@ int xdo_set_window_property(const xdo_t *xdo, Window wid, const char *property,
   
   char netwm_property[256] = "_NET_";
   int ret = 0;
-  strncat(netwm_property, property, strlen(property));
+  strcat(netwm_property, property);
 
   // Change the property
   ret = XChangeProperty(xdo->xdpy, wid, 
@@ -977,7 +996,7 @@ int xdo_enter_text_window(const xdo_t *xdo, Window window, const char *string, u
     }
     _xdo_charcodemap_from_char(xdo, &key);
     if (key.code == 0 && key.symbol == NoSymbol) {
-      fprintf(stderr, "I don't what key produces '%lc', skipping.\n",
+      fprintf(stderr, "I don't know which key produces '%lc', skipping.\n",
               key.key);
       continue;
     } else {
@@ -1260,6 +1279,14 @@ static KeySym _xdo_keysym_from_char(const xdo_t *xdo, wchar_t key) {
 static void _xdo_charcodemap_from_char(const xdo_t *xdo, charcodemap_t *key) {
   KeySym keysym = _xdo_keysym_from_char(xdo, key->key);
   _xdo_charcodemap_from_keysym(xdo, key, keysym);
+
+  /* If the character is an uppercase character within the Basic Latin or Latin-1 code block,
+   * then sending the capital character keycode will not work.
+   * We have to also send the shift modifier.
+   * There are only three ranges of capital letters to worry about */
+  if ((key->key >= 0x41 && key->key <= 0x5A) || (key->key >= 0xC0 && key->key <= 0xD6) || (key->key >= 0xD8 && key->key <= 0xDE)) {
+    key->modmask = ShiftMask;
+  }
 }
 
 static void _xdo_charcodemap_from_keysym(const xdo_t *xdo, charcodemap_t *key, KeySym keysym) {
@@ -1491,8 +1518,10 @@ int _xdo_ewmh_is_supported(const xdo_t *xdo, const char *feature) {
 
   results = (Atom *) xdo_get_window_property_by_atom(xdo, root, request, &nitems, &type, &size);
   for (i = 0L; i < nitems; i++) {
-    if (results[i] == feature_atom)
+    if (results[i] == feature_atom) {
+      free(results);
       return True;
+    }
   }
   free(results);
 
@@ -1790,18 +1819,21 @@ int xdo_get_desktop_viewport(const xdo_t *xdo, int *x_ret, int *y_ret) {
             "Got unexpected type returned from _NET_DESKTOP_VIEWPORT."
             " Expected CARDINAL, got %s\n",
             XGetAtomName(xdo->xdpy, type));
+    free(data);
     return XDO_ERROR;
   }
 
   if (nitems != 2) {
     fprintf(stderr, "Expected 2 items for _NET_DESKTOP_VIEWPORT, got %ld\n",
             nitems);
+    free(data);
     return XDO_ERROR;
   }
 
   int *viewport_data = (int *)data;
   *x_ret = viewport_data[0];
   *y_ret = viewport_data[1];
+  free(data);
 
   return XDO_SUCCESS;
 }
@@ -1841,6 +1873,29 @@ int xdo_close_window(const xdo_t *xdo, Window window) {
   return _is_success("XDestroyWindow", ret == 0, xdo);
 }
 
+int xdo_quit_window(const xdo_t *xdo, Window window) {
+  XEvent xev;
+  int ret;
+  Window root = RootWindow(xdo->xdpy, 0);
+
+  memset(&xev, 0, sizeof(xev));
+  xev.type = ClientMessage;
+  xev.xclient.serial = 0;
+  xev.xclient.send_event = True;
+  xev.xclient.display = xdo->xdpy;
+  xev.xclient.window = window;
+  xev.xclient.message_type = XInternAtom(xdo->xdpy, "_NET_CLOSE_WINDOW", False);
+  xev.xclient.format = 32;
+
+  ret = XSendEvent(xdo->xdpy, root, False,
+                   SubstructureNotifyMask | SubstructureRedirectMask,
+                   &xev);
+
+  /* XXX: XSendEvent returns 0 on conversion failure, nonzero otherwise.
+   * Manpage says it will only generate BadWindow or BadValue errors */
+  return _is_success("XSendEvent[_NET_CLOSE_WINDOW]", ret == 0, xdo);
+}
+
 int xdo_get_window_name(const xdo_t *xdo, Window window, 
                         unsigned char **name_ret, int *name_len_ret,
                         int *name_type) {
@@ -1879,6 +1934,34 @@ int xdo_get_window_name(const xdo_t *xdo, Window window,
   return 0;
 }
 
+int xdo_get_window_classname(const xdo_t *xdo, Window window, unsigned char **name_ret) {
+  XClassHint classhint;
+  XGetClassHint(xdo->xdpy, window, &classhint);
+  XFree(classhint.res_name);
+  *name_ret = (unsigned char*) classhint.res_class;
+  return 0;
+}
+
+int xdo_window_state(xdo_t *xdo, Window window, unsigned long action, const char *property) {
+  int ret;
+  XEvent xev;
+  Window root = RootWindow(xdo->xdpy, 0);
+
+  memset(&xev, 0, sizeof(xev));
+  xev.xclient.type = ClientMessage;
+  xev.xclient.serial = 0;
+  xev.xclient.send_event = True;
+  xev.xclient.message_type = XInternAtom(xdo->xdpy, "_NET_WM_STATE", False);
+  xev.xclient.window = window;
+  xev.xclient.format = 32;
+  xev.xclient.data.l[0] = action;
+  xev.xclient.data.l[1] = XInternAtom(xdo->xdpy, property, False);
+
+  ret = XSendEvent(xdo->xdpy, root, False,
+                   SubstructureNotifyMask | SubstructureRedirectMask, &xev);
+  return _is_success("XSendEvent[EWMH:_NET_WM_STATE]", ret == 0, xdo);
+}
+
 int xdo_minimize_window(const xdo_t *xdo, Window window) {
   int ret;
   int screen;
@@ -1953,3 +2036,24 @@ int xdo_get_viewport_dimensions(xdo_t *xdo, unsigned int *width,
     return xdo_get_window_size(xdo, root, width, height);
   }
 }
+
+static int appears_to_be_wayland(Display *xdpy) {
+  // XWayland leaks its presence in two extensions (at time of writing, August 2021)
+  // First: the name of input devices "xwayland-pointer"
+  // Second: in the Vendor string of XFree86-VidModeExtension
+
+  int count;
+  XDeviceInfo *devices = XListInputDevices(xdpy, &count);
+  for (int i = 0; i < count; i++) {
+    // fprintf(stderr, "Device %d: %s\n", i, devices[i].name);
+    // If the input device name starts with "xwayland-", 
+    // there's a good chance we're running on XWayland.
+    if (strstr(devices[i].name, "xwayland-") == devices[i].name) {
+      XFreeDeviceList(devices);
+      return 1; // Running on wayland
+    }
+  }
+  XFreeDeviceList(devices);
+
+  return 0;
+}
\ No newline at end of file
